#lang racket
(require "assembly.rkt"
         test-engine/racket-tests)

;; An decl is (cons sym list-of-declet)

;; An declet is either
;;  - symbol
;;  - number

;; A dictionary is a list-of-name
;; A name is (make-name sym num)
(define-struct name (s val) #:transparent)

(provide collect-names
         subst-and-assemble)

;; collect-names : list-of-decl number -> dictionary
;;  Takes a list of declarations and produces a dictionary mapping 
;;  names to numbers.
(define (collect-names l index)
  (cond
    [(empty? l) empty]
    [(cons? l) 
     (define d (first l))
     (cond
       [(eq? (first d) 'label) ; d is '(label sym) for some sym
        (cons (make-name (second d) index)
              (collect-names (rest l) index))]
       [(eq? (first d) 'const) ; d is '(const sym num)
        (cons (make-name (second d) (third d))
              (collect-names (rest l) index))]
       [(eq? (first d) 'data) ; d is '(data sym declet ...)
        (define data-size (- (length d) 2)) ; the number of declets
        (cons (make-name (second d) index)
              (collect-names (rest l) (+ index data-size)))]
       [else
        (collect-names (rest l) (+ index 1))])]))

(check-expect (collect-names empty 0)
              empty)
(check-expect (collect-names '((ld R0 R0) (halt)) 0)
              empty)
(check-expect (collect-names '((ld R0 R0) (label MIDDLE) (halt)) 0)
              (list (make-name 'MIDDLE 1)))
(check-expect (collect-names '((ld R0 R0) (label MIDDLE) (label END) (halt)) 0)
              (list (make-name 'MIDDLE 1)
                    (make-name 'END 1)))
(check-expect (collect-names '((ld R0 R0) (label MIDDLE) (halt)) 7)
              (list (make-name 'MIDDLE 8)))
(check-expect (collect-names '((ld R0 R0) (const MIDDLE 9) (halt)) 7)
              (list (make-name 'MIDDLE 9)))
(check-expect (collect-names '((ld R0 R0) (data MIDDLE 9 10 11) (halt)) 7)
              (list (make-name 'MIDDLE 8)))
(check-expect (collect-names '((ld R0 R0) (data MIDDLE 9 10 11) (label END) (halt)) 7)
              (list (make-name 'MIDDLE 8)
                    (make-name 'END 11)))

;; find-name : declet dictionary -> declet
;;  Takes a symbol or number and a dictionary and returns a symbol or number.
;;  The result is the same as the given symbol or number, unless the symbol 
;;  has a value in the dictionary, in which case the symbol’s value is returned,
;;  instead.
(define (find-name dlet dict)
  (cond
    [(empty? dict) dlet]
    [(cons? dict) 
     (define n (first dict))
     (if (eq? dlet (name-s n))
         (name-val n)
         (find-name dlet (rest dict)))]))
  
(check-expect (find-name 'MIDDLE empty) 'MIDDLE)
(check-expect (find-name 42 empty) 42)
(check-expect (find-name 42 (list (make-name 'MIDDLE 8))) 42)
(check-expect (find-name 'MIDDLE (list (make-name 'MIDDLE 8))) 8)
(check-expect (find-name 'MIDDLE (list (make-name 'FIRST 0)
                                       (make-name 'MIDDLE 8))) 
              8)

;; subst : decl dictionary -> decl
;;  Takes a declaration and a dictionary and returns a declaration where 
;;  each symbol that has a value in the dictionary is replaced by its value.
(define (subst d dict)
  (map (lambda (dlet) (find-name dlet dict))
       d))

(check-expect (subst '(ldi R0 MIDDLE) empty)
              '(ldi R0 MIDDLE))
(check-expect (subst '(ldi R0 MIDDLE) (list (make-name 'MIDDLE 9)))
              '(ldi R0 9))
(check-expect (subst '(ldi R0 END) (list (make-name 'MIDDLE 9)
                                         (make-name 'END 10)))
              '(ldi R0 10))

;; subst-and-assemble : list-of-decl dictionary -> list-of-number
;;  Takes a list of declarations and a dictionary and returns a list of
;;  numbers (for machine codes). For each declaration, the result should
;;  contain a corresponding machine code (if the declaration generates a 
;;  machine code) that is either the data value for a data declaration or 
;;  the result of `assemble' for an instruction.
(define (subst-and-assemble l dict)
  (cond
    [(empty? l) empty]
    [(cons? l)
     (define d (first l))
     (cond
       [(eq? 'const (first d))
        (subst-and-assemble (rest l) dict)]
       [(eq? 'label (first d))
        (subst-and-assemble (rest l) dict)]
       [(eq? 'data (first d)) ; d is '(data sym declet ...)
        (append (rest (rest (subst d dict)))
                (subst-and-assemble (rest l) dict))]
       [else
        (cons (assemble (subst d dict))
              (subst-and-assemble (rest l) dict))])]))

(define dict1 (list (make-name 'MIDDLE 9)
                    (make-name 'START 1)
                    (make-name 'END 10)))

(check-expect (subst-and-assemble empty empty)
              empty)
(check-expect (subst-and-assemble (list '(ldi R0 END)) dict1)
              (list (assemble '(ldi R0 10))))
(check-expect (subst-and-assemble (list '(ldi R0 END)
                                        '(label START))
                                  dict1)
              (list (assemble '(ldi R0 10))))
(check-expect (subst-and-assemble (list '(ldi R0 END)
                                        '(const START 1))
                                  dict1)
              (list (assemble '(ldi R0 10))))
(check-expect (subst-and-assemble (list '(label START)
                                        '(ldi R0 END))
                                  dict1)
              (list (assemble '(ldi R0 10))))
(check-expect (subst-and-assemble (list '(halt)
                                        '(label START)
                                        '(ldi R0 END))
                                  dict1)
              (list (assemble '(halt))
                    (assemble '(ldi R0 10))))
(check-expect (subst-and-assemble (list '(halt)
                                        '(data START 0 1 2 3 END 5 6)
                                        '(ldi R0 END))
                                  dict1)
              (list (assemble '(halt))
                    0 1 2 3 10 5 6
                    (assemble '(ldi R0 10))))


(test-silence true)
(test)
